#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import os
import re
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_pa_func()
    test.test_cr_ver_del_ver_default()
    test.test_upd_ver_del()
    utils_libs.report_results('test_gbp_pa_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_pa_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_pa_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_pa_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_ACTION FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_act'

    def cleanup(self, cfgobj, uuid_name, tc_name=''):
        if tc_name != '':
            self._log.info('%s FAILED' % (tc_name))
        if isinstance(cfgobj, str):
            cfgobj = [cfgobj]
        if isinstance(uuid_name, str):
            uuid_name = [uuid_name]
        for obj, _id in zip(cfgobj, uuid_name):
            if self.gbpcfg.gbp_policy_cfg_all(0, obj, _id):
                self._log.info(
                    'Success in Clean-up/Delete of Policy Object %s\n' %
                    (obj))
            else:
                self._log.info(
                    'Failed to Clean-up/Delete of Policy Object %s\n' %
                    (obj))
        os._exit(1)

    def test_cr_ver_del_ver_default(self, rep_cr=0, rep_del=0):

        if rep_cr == 0 and rep_del == 0:
            self._log.info(
                "\n###################################################\n"
                "TESTCASE_GBP_PA_FUNC_1: CREATE/VERIFY/DELETE/VERIFY a "
                "POLICY ACTION with DEFAULT ATTR VALUE\n"
                "TEST_STEPS:\n"
                "Create Policy Action Object,default params\n"
                "Verify the attributes & value, show & list cmds\n"
                "Delete Policy Action using Name\n"
                "Verify that PA has got deleted, show & list cmds\n"
                "Recreate Policy Action Object inorder to test Delete "
                "using UUID\n"
                "Delete using UUID\n"
                "Verify that PA has got deleted, show & list cmds\n"
                "###################################################\n")

        # Testcase work-flow starts
        self._log.info(
            '\n## Step 1: Create Action with default attrib values##\n')
        act_uuid = self.gbpcfg.gbp_action_config(1, self.act_name)
        if act_uuid == 0:
            self._log.info("# Step 1: Create Action == Failed")
            return 0
        if self.gbpverify.gbp_action_verify(0, self.act_name, act_uuid) == 0:
            self._log.info(
                "# Step 2A: Verify Action using -list option == Failed")
            return 0
        if self.gbpverify.gbp_action_verify(
                1,
                self.act_name,
                id=act_uuid,
                action_type='allow',
                shared='False') == 0:
            self._log.info(
                "# Step 2B: Verify Action using -show option == Failed")
            return 0
        ######
        self._log.info('\n## Step 3: Delete Action using name ##\n')
        if self.gbpcfg.gbp_action_config(0, self.act_name) == 0:
            self._log.info("# Step 3: Delete Action using Name == Failed")
            return 0
        if self.gbpverify.gbp_action_verify(0, self.act_name, act_uuid) != 0:
            self._log.info(
                "\n## Step 3A: Verify Action is Deleted using -list option "
                "== Failed")
            return 0
        if self.gbpverify.gbp_action_verify(
                1,
                self.act_name,
                id=act_uuid,
                action_type='allow',
                shared='False') != 0:
            self._log.info(
                "\n## Step 3B: Verify Action is Deleted using -show option "
                "== Failed")
            return 0

        act_uuid = self.gbpcfg.gbp_action_config(1, self.act_name)
        if act_uuid:
            self._log.info(
                "Step 4: Re-created a Policy Action with default inorder "
                "to delete with ID")
            self._log.info('\n## Step 5: Delete Action using UUID ##\n')
            if self.gbpcfg.gbp_action_config(0, act_uuid) == 0:
                self._log.info(
                    "\n## Step 5: Delete Action using UUID == Failed")
                return 0
            if self.gbpverify.gbp_action_verify(
                    0, act_uuid, self.act_name) != 0:
                self._log.info(
                    "\n## Step 5A: Verify Action is Deleted using -list "
                    "option == Failed")
                return 0
            if self.gbpverify.gbp_action_verify(
                    1,
                    act_uuid,
                    name=self.act_name,
                    action_type='allow',
                    shared='False') != 0:
                self._log.info(
                    "\n## Step 5B: Verify Action is Deleted using -show "
                    "option == Failed")
                return 0
            self._log.info(
                "\n## Step 5: Delete of Policy Action using UUID == Passed")
        else:
            self._log.info(
                "\n## Step 6: Recreate of Policy Action using Default "
                "== Failed")
            return 0
        if rep_cr == 0 and rep_del == 0:
            self._log.info("\n## TESTCASE_GBP_PA_FUNC_1: PASSED")
        return 1

    def test_upd_ver_del(self):
        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PA_FUNC_2: UPDATE/VERIFY/DELETE EDITABLE ATTRIBs "
            "of a POLICY ACTION \n"
            "TEST_STEPS::\n"
            "Create Policy Action using Default param values\n"
            "Update the Polciy Action's editable params\n"
            "Verify the Policy Action's attributes & values, show & list "
            "cmds\n"
            "Delete the Policy Action\n"
            "Verify Policy Action successfully deleted\n"
            "###################################################\n")

        # Testcase work-flow starts
        self._log.info(
            '\n##Step 1: Create Action with default attrib vals ##\n')
        act_uuid = self.gbpcfg.gbp_action_config(1, self.act_name)
        if act_uuid == 0:
            self._log.info("## Step 1: Create Action == Failed")
            return 0
        self._log.info(
            "\n## Step 1A: Creating a Service Chain Spec to be used for "
            "UPdating Polic Action")
        spec_cr_cmd = ('gbp servicechain-spec-create demo_spec | grep id | '
                       'head -1')
        cmd_out = commands.getoutput(spec_cr_cmd)
        spec_id = re.search("\\bid\\b\s+\| (.*) \|", cmd_out, re.I).group(1)
        self._log.info(
            '\n##Step 2: Update Policy Action Attributes name and '
            'action_value##\n')
        if self.gbpcfg.gbp_action_config(
                2,
                act_uuid,
                name='grppol_act',
                action_value=spec_id) == 0:
            self._log.info(
                "\n##Step 2: Updating Policy Action's Attributes name "
                "& action_value, Failed")
            return 0

        if self.gbpverify.gbp_action_verify(
                0, 'grppol_act', act_uuid, spec_id) == 0:
            self._log.info(
                "\n## Step 2A: Verify Policy Action Updated Attributes "
                "using -list option == Failed")
            return 0
        if self.gbpverify.gbp_action_verify(
                1,
                'grppol_act',
                id=act_uuid,
                action_type='allow',
                shared='False',
                action_value=spec_id) == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Action Updated Attributes "
                "using -show option == Failed")
            return 0
        if self.gbpcfg.gbp_action_config(0, act_uuid) == 0:
            self._log.info("## Step 3: Delete Action using Name == Failed")
            return 0
        self._log.info("\n## Step 3A: Now delete the service chain spec")
        spec_del_cmd = 'gbp servicechain-spec-delete %s' % (spec_id)
        cmd_out = commands.getoutput(spec_del_cmd)
        if self.gbpverify.gbp_action_verify(
                1,
                'grppol_act',
                id=act_uuid,
                action_type='allow',
                shared='False') != 0:
            self._log.info(
                "\n## Step 3B: Verify Action is Deleted using -show "
                "option == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PA_FUNC_2: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import neutron.common  # noqa

from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_group_proxy_extension as test_gp_ext)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_resource_mapping)


class ResourceMappingProxyGroupGBPTestCase(
        test_resource_mapping.ResourceMappingTestCase):

    def setUp(self):
        config.cfg.CONF.set_override('extension_drivers',
                                     ['proxy_group'],
                                     group='group_policy')
        super(ResourceMappingProxyGroupGBPTestCase, self).setUp()

    def _get_l3p_from_ptg(self, ptg):
        l2p = self.show_l2_policy(ptg['l2_policy_id'])['l2_policy']
        return self.show_l3_policy(l2p['l3_policy_id'])['l3_policy']

    def _is_router_attached_to_any_subnet(self, router_ports, subnet_ids):
        subnet_ids = set(subnet_ids)
        for port in router_ports:
            if any([x for x in port['fixed_ips']
                    if x['subnet_id'] in subnet_ids]):
                return True
        return False

    def _is_router_attached_to_all_subnets(self, router_ports, subnet_ids):
        subnet_ids = set(subnet_ids)
        for port in router_ports:
            for subnet_id in [x['subnet_id'] for x in port['fixed_ips'] if
                              x['subnet_id'] in subnet_ids]:
                subnet_ids.remove(subnet_id)
        return not bool(subnet_ids)

    def _ptg_to_cidrs(self, ptg):
        cidrs = []
        for sub_id in ptg['subnets']:
            cidrs.append(
                self._get_object(
                    'subnets', sub_id, self.api)['subnet']['cidr'])
        return cidrs

    def _are_subnets_same(self, ptg1, ptg2):
        cidrs1 = netaddr.IPSet(iterable=self._ptg_to_cidrs(ptg1))
        cidrs2 = netaddr.IPSet(iterable=self._ptg_to_cidrs(ptg2))
        return cidrs1 == cidrs2

    def _are_subnets_disjoint(self, ptg1, ptg2):
        cidrs1 = netaddr.IPSet(iterable=self._ptg_to_cidrs(ptg1))
        cidrs2 = netaddr.IPSet(iterable=self._ptg_to_cidrs(ptg2))
        return len(cidrs1 & cidrs2) == 0

    def _ptg_gateway_pt(self, ptg):
        pt_list = self._list('policy_targets',
            query_params='proxy_gateway=True&policy_target_group_id=%s' %
                         ptg['id'])
        if pt_list['policy_targets']:
            return pt_list['policy_targets'][0]

    def _get_pt_ip(self, pt):
        port = self._get_object('ports', pt['port_id'], self.api)['port']
        return port['fixed_ips'][0]['ip_address']

    def _get_ptg_subnets(self, ptg):
        return [self._get_object('subnets', x, self.api)['subnet']
                for x in ptg['subnets']]

    def _are_cidr_routed_in_ptg_via_pt(self, ptg, pt, cidrs):
        nexthop = self._get_pt_ip(pt)
        subnets = self._get_ptg_subnets(ptg)
        expected_routes = set((destination, nexthop) for destination in cidrs)
        for subnet in subnets:
            expected_routes = (
                expected_routes - set((x['destination'], x['nexthop'])
                                      for x in subnet['host_routes']))
        return not expected_routes

    def _are_cidr_routed_in_router_via_pt(self, router, pt, cidrs):
        nexthop = self._get_pt_ip(pt)
        expected_routes = set((destination, nexthop) for destination in cidrs)
        routes = set((x['destination'], x['nexthop'])
                     for x in router['routes'])
        return not expected_routes - routes

    def _verify_correct_proxy_chain(self, original_ptg_id,
                                    expected_chain_length=None):
        original_ptg = self.show_policy_target_group(
            original_ptg_id)['policy_target_group']
        l3p = self._get_l3p_from_ptg(original_ptg)
        router = self._get_object(
            'routers', l3p['routers'][0], self.ext_api)['router']
        router_ports = self._list(
            'ports', query_params='device_id=%s' % router['id'])['ports']
        cumulative_cidrs = set()
        curr = original_ptg
        chain_length = 1
        while curr['proxy_group_id']:
            chain_length += 1
            proxy = self.show_policy_target_group(
                curr['proxy_group_id'])['policy_target_group']
            # Current PTG is not proxied, therefore it mustn't be attached to
            # the L3P router
            self.assertFalse(
                self._is_router_attached_to_any_subnet(
                    router_ports, curr),
                "Some ports are still attached to the PTG "
                "subnets:\nports:\n%s\nsubnets:\n%s" %
                (router_ports, curr['subnets']))
            # If L3 proxy, all subnets must be on different CIDRs
            if proxy['proxy_type'] == 'l2':
                self.assertTrue(
                    self._are_subnets_same(proxy, curr),
                    "PTG cidrs don't overlap on L2 proxy"
                    ":\nproxy_cidrs:\n%s\nproxied_cidrs:\n%s\n" %
                    (self._ptg_to_cidrs(proxy), (self._ptg_to_cidrs(curr))))
            else:
                self.assertTrue(
                    self._are_subnets_disjoint(proxy, curr),
                    "PTG cidrs overlap on L3 proxy"
                    ":\nproxy_cidrs:\n%s\nproxied_cidrs:\n%s\n" %
                    (self._ptg_to_cidrs(proxy), (self._ptg_to_cidrs(curr))))
            # Check that routes are correctly set
            curr_gateway_pt = self._ptg_gateway_pt(curr)
            if curr_gateway_pt:
                # Cumulated subnets are routed in current PTG subnets
                self.assertTrue(
                    self._are_cidr_routed_in_ptg_via_pt(curr, curr_gateway_pt,
                                                        cumulative_cidrs),
                    "PTG is not routing all the chained subnets:"
                    "\nPTG subnets:\n%s\nExpected routed CIDRs:%s\nExpected "
                    "nexthop:%s" % (self._get_ptg_subnets(curr),
                                    cumulative_cidrs,
                                    self._get_pt_ip(curr_gateway_pt)))
            cumulative_cidrs.update(self._ptg_to_cidrs(curr))
            curr = proxy
        # The last PTG of the proxy chain is fully connected to the router
        self.assertTrue(
            self._is_router_attached_to_all_subnets(
                router_ports, curr['subnets']),
            "Some ports are not attached to the PTG "
            "subnets:\nports:\n%s\nsubnets:\n%s" % (router_ports,
                                                    curr['subnets']))
        # Verify last PTG routes
        curr_gateway_pt = self._ptg_gateway_pt(curr)
        if curr_gateway_pt:
            # Cumulated subnets are routed in current PTG subnets
            self.assertTrue(
                self._are_cidr_routed_in_ptg_via_pt(curr, curr_gateway_pt,
                                                    cumulative_cidrs),
                "PTG is not routing all the chained subnets:"
                "\nPTG subnets:\n%s\nExpected routed CIDRs:%s\nExpected "
                "nexthop:%s" % (self._get_ptg_subnets(curr),
                                cumulative_cidrs,
                                self._get_pt_ip(curr_gateway_pt)))
            # Cumulated subnets are routed in L3P
            self.assertTrue(
                self._are_cidr_routed_in_router_via_pt(
                    router, curr_gateway_pt, cumulative_cidrs),
                "Router is not routing all the chained subnets:"
                "\nRouter:\n%s\nExpected routed CIDRs:%s\nExpected "
                "nexthop:%s" % (router, cumulative_cidrs,
                                self._get_pt_ip(curr_gateway_pt)))

        if expected_chain_length:
            self.assertEqual(expected_chain_length, chain_length)


class TestProxyGroupSubnetPrefixRMD(ResourceMappingProxyGroupGBPTestCase):

    def setUp(self):
        config.cfg.CONF.set_override(
                'default_proxy_subnet_prefix_length', '26',
                group='group_policy_proxy_group')
        config.cfg.CONF.set_override(
                'default_proxy_ip_pool', '192.168.1.0/24',
                group='group_policy_proxy_group')
        super(TestProxyGroupSubnetPrefixRMD, self).setUp()

    def test_proxy_group_updated_prefix_length(self):
        l3p = self.create_l3_policy(ip_pool='11.0.0.0/8')['l3_policy']
        self.assertEqual('192.168.1.0/24', l3p['proxy_ip_pool'])
        self.assertEqual(26, l3p['proxy_subnet_prefix_length'])

        l2p = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        l2p2 = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg_proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            l2_policy_id=l2p2['id'])['policy_target_group']

        subnet = self._get_object('subnets', ptg_proxy['subnets'][0],
                                  self.api)['subnet']
        self.assertEqual(str(l3p['proxy_subnet_prefix_length']),
                         subnet['cidr'].split('/')[1])


class TestProxyGroupRMD(ResourceMappingProxyGroupGBPTestCase,
                        test_gp_ext.ExtensionDriverTestCaseMixin):

    def test_proxy_group_extension(self):
        l3p = self.create_l3_policy(ip_pool='11.0.0.0/8')['l3_policy']
        self.assertEqual('192.168.0.0/16', l3p['proxy_ip_pool'])
        self.assertEqual(28, l3p['proxy_subnet_prefix_length'])

        l2p = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        self.assertIsNone(ptg['proxy_group_id'])
        self.assertIsNone(ptg['proxied_group_id'])
        self.assertIsNone(ptg['proxy_type'])

        # Verify Default L3P pool mapping on show
        l3p = self.show_l3_policy(l3p['id'])['l3_policy']
        self.assertEqual('192.168.0.0/16', l3p['proxy_ip_pool'])
        self.assertEqual(28, l3p['proxy_subnet_prefix_length'])

        l2p2 = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg_proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'],
            l2_policy_id=l2p2['id'])['policy_target_group']
        self.assertIsNone(ptg_proxy['proxy_group_id'])
        self.assertEqual(ptg['id'], ptg_proxy['proxied_group_id'])
        self.assertEqual('l3', ptg_proxy['proxy_type'])

        subnet = self._get_object('subnets', ptg_proxy['subnets'][0],
                                  self.api)['subnet']
        self.assertEqual(str(l3p['proxy_subnet_prefix_length']),
                         subnet['cidr'].split('/')[1])

        # Verify relationship added
        ptg = self.show_policy_target_group(ptg['id'])['policy_target_group']
        self.assertEqual(ptg_proxy['id'], ptg['proxy_group_id'])
        self.assertIsNone(ptg['proxied_group_id'])

        pt = self.create_policy_target(
            policy_target_group_id=ptg_proxy['id'])['policy_target']
        self.assertFalse(pt['proxy_gateway'])
        pt = self.create_policy_target(
            policy_target_group_id=ptg_proxy['id'],
            proxy_gateway=True)['policy_target']
        self.assertTrue(pt['proxy_gateway'])
        pt = self.show_policy_target(pt['id'])['policy_target']
        self.assertTrue(pt['proxy_gateway'])


class TestPolicyTarget(ResourceMappingProxyGroupGBPTestCase,
                       test_resource_mapping.TestPolicyTarget):

    def test_proxy_gateway(self):
        # Verify that L3 Proxy is correctly stitched
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l3')['policy_target_group']
        self.create_policy_target(policy_target_group_id=proxy['id'],
                                  proxy_gateway=True)
        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=2)

        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy['id'],
            proxy_type='l3')['policy_target_group']
        self.create_policy_target(policy_target_group_id=proxy2['id'],
                                  proxy_gateway=True)

        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=3)

    def test_group_default_gateway(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l3')

        pt = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            group_default_gateway=True)['policy_target']
        port = self._get_object('ports', pt['port_id'], self.api)['port']
        self.assertGreater(len(port['fixed_ips']), 0)
        for fixed_ip in port['fixed_ips']:
            subnet = self._get_object('subnets', fixed_ip['subnet_id'],
                                      self.api)['subnet']
            self.assertEqual(subnet['gateway_ip'], fixed_ip['ip_address'])

    def test_group_default_gateway_fails(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target(policy_target_group_id=ptg['id'],
                                  group_default_gateway=True,
                                  expected_res_status=409)

    def test_proxy_gateway_deleted(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l3')['policy_target_group']
        pt = self.create_policy_target(
            policy_target_group_id=proxy['id'],
            proxy_gateway=True)['policy_target']
        nexthop = self._get_pt_ip(pt)
        l3p = self._get_l3p_from_ptg(ptg)
        router = self._get_object(
            'routers', l3p['routers'][0], self.ext_api)['router']
        subnet = self._get_object(
            'subnets', proxy['subnets'][0], self.api)['subnet']

        self.assertTrue(
            any([x for x in router['routes'] if x['nexthop'] == nexthop]),
            "No routes in router with "
            "nexthop %s:\n%s" % (nexthop, router['routes']))
        self.assertTrue(
            any([x for x in subnet['host_routes'] if x['nexthop'] == nexthop]),
            "No routes in PTG with "
            "nexthop %s:\n%s" % (nexthop, subnet['host_routes']))

        self.delete_policy_target(pt['id'])
        # Verify routes deleted for that PT
        router = self._get_object(
            'routers', l3p['routers'][0], self.ext_api)['router']
        subnet = self._get_object(
            'subnets', ptg['subnets'][0], self.api)['subnet']
        self.assertFalse(
            any([x for x in router['routes'] if x['nexthop'] == nexthop]),
            "Some routes still remains in router with "
            "nexthop %s:\n%s" % (nexthop, router['routes']))
        self.assertFalse(
            any([x for x in subnet['host_routes'] if x['nexthop'] == nexthop]),
            "Some routes still remains in subnet with "
            "nexthop %s:\n%s" % (nexthop, subnet['host_routes']))

    def test_implicit_port_lifecycle(self):
        super(TestPolicyTarget, self).test_implicit_port_lifecycle(
            proxy_ip_pool='182.169.0.0/16')

    def test_weird_port_extra_attributes_ignored(self):
        super(TestPolicyTarget, self).test_weird_port_extra_attributes_ignored(
            extra={'proxy_gateway': False, 'group_default_gateway': False})

    def test_port_extra_attributes(self):
        super(TestPolicyTarget, self).test_port_extra_attributes(
            extra={'proxy_gateway': False, 'group_default_gateway': False})

    def test_port_extra_attributes_fixed_ips(self):
        super(TestPolicyTarget, self).test_port_extra_attributes_fixed_ips(
            extra={'proxy_gateway': False, 'group_default_gateway': False})

    def test_port_extra_attributes_implicit(self):
        super(TestPolicyTarget, self).test_port_extra_attributes_implicit(
            extra={'proxy_gateway': False, 'group_default_gateway': False})


class TestPolicyTargetGroup(ResourceMappingProxyGroupGBPTestCase,
                            test_resource_mapping.TestPolicyTargetGroup):

    def test_l3_proxy_stitching(self):
        # Verify that L3 Proxy is correctly stitched
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target_group(proxied_group_id=ptg['id'],
                                        proxy_type='l3')
        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=2)

    def test_l2_proxy_stitching(self):
        # Verify that L2 Proxy is correctly stitched
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target_group(proxied_group_id=ptg['id'],
                                        proxy_type='l2')
        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=2)

    def test_multi_type_stitcing(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy1 = self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l3')['policy_target_group']
        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy1['id'],
            proxy_type='l2')['policy_target_group']
        self.assertEqual('l2', proxy2['proxy_type'])
        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=3)

        # Add L3 proxy in front of the chain
        proxy3 = self.create_policy_target_group(
            proxied_group_id=proxy2['id'],
            proxy_type='l3')['policy_target_group']
        self.assertEqual('l3', proxy3['proxy_type'])
        self._verify_correct_proxy_chain(ptg['id'], expected_chain_length=4)

    def test_multi_type_stitcing_delete_head(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        proxy1 = self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l3')['policy_target_group']
        proxy2 = self.create_policy_target_group(
            proxied_group_id=proxy1['id'],
            proxy_type='l2')['policy_target_group']

        # Add L3 proxy in front of the chain
        proxy3 = self.create_policy_target_group(
            proxied_group_id=proxy2['id'],
            proxy_type='l3')['policy_target_group']
        self.delete_policy_target_group(ptg['id'], expected_res_status=204)
        self.show_policy_target_group(proxy1['id'], expected_res_status=404)
        self.show_policy_target_group(proxy2['id'], expected_res_status=404)
        self.show_policy_target_group(proxy3['id'], expected_res_status=404)

    def test_proxy_creation_different_l3p(self):
        l3p1 = self.create_l3_policy()['l3_policy']
        l2p1 = self.create_l2_policy(l3_policy_id=l3p1['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p1['id'])['policy_target_group']

        l3p2 = self.create_l3_policy(
            ip_pool='11.0.0.0/8', proxy_ip_pool='192.169.0.0/26')['l3_policy']
        l2p2 = self.create_l2_policy(l3_policy_id=l3p2['id'])['l2_policy']
        res = self.create_policy_target_group(
            l2_policy_id=l2p2['id'], proxied_group_id=ptg['id'],
            expected_res_status=400)
        self.assertEqual('InvalidProxiedGroupL3P', res['NeutronError']['type'])

    def test_l2_proxy_creation_same_l2p(self):
        l3p1 = self.create_l3_policy()['l3_policy']
        l2p1 = self.create_l2_policy(l3_policy_id=l3p1['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p1['id'])['policy_target_group']
        res = self.create_policy_target_group(
            l2_policy_id=l2p1['id'], proxied_group_id=ptg['id'],
            proxy_type='l2', expected_res_status=400)
        self.assertEqual('InvalidProxiedGroupL2P', res['NeutronError']['type'])

    def test_l3_proxy_creation_same_l2p(self):
        l3p1 = self.create_l3_policy()['l3_policy']
        l2p1 = self.create_l2_policy(l3_policy_id=l3p1['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p1['id'])['policy_target_group']
        self.create_policy_target_group(
            l2_policy_id=l2p1['id'], proxied_group_id=ptg['id'],
            proxy_type='l3', expected_res_status=201)

    def test_proxy_creation_implicit_l2p(self):
        l3p = self.create_l3_policy(
            name='nondefault', ip_pool='11.0.0.0/8',
            proxy_ip_pool='193.168.0.0/16')['l3_policy']
        l2p = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']

        # Create a normal group with implicit L2P, will be created in the
        # default context
        group_noproxy = self.create_policy_target_group(
            l2_policy_id=None)['policy_target_group']
        l2p = self.show_l2_policy(group_noproxy['l2_policy_id'])['l2_policy']
        self.assertNotEqual(l3p['id'], l2p['l3_policy_id'])

        # Create a proxy group with implicit L2P, will be created in the
        # default context
        group_proxy = self.create_policy_target_group(
            l2_policy_id=None,
            proxied_group_id=ptg['id'])['policy_target_group']
        l2p = self.show_l2_policy(group_proxy['l2_policy_id'])['l2_policy']
        self.assertEqual(l3p['id'], l2p['l3_policy_id'])


class TestL2Policy(ResourceMappingProxyGroupGBPTestCase,
                   test_resource_mapping.TestL2Policy):

    def test_l3p_update_rejected(self):
        super(TestL2Policy, self).test_l3p_update_rejected(
            proxy_ip_pool='182.169.0.0/16')


class TestL3Policy(ResourceMappingProxyGroupGBPTestCase,
                   test_resource_mapping.TestL3Policy):

    def test_implicit_l3p_proxy_pool(self):
        default_proxy_pool = '192.168.0.0/16'
        default_proxy_subnet_prefix_length = 28
        l2p = self.create_l2_policy()['l2_policy']
        l3p = self.show_l3_policy(l2p['l3_policy_id'])['l3_policy']
        self.assertEqual(default_proxy_pool, l3p['proxy_ip_pool'])
        self.assertEqual(default_proxy_subnet_prefix_length,
                         l3p['proxy_subnet_prefix_length'])


class TestPolicyRuleSet(ResourceMappingProxyGroupGBPTestCase,
                        test_resource_mapping.TestPolicyRuleSet):
    pass


class TestPolicyAction(ResourceMappingProxyGroupGBPTestCase,
                       test_resource_mapping.TestPolicyAction):
    pass


class TestPolicyRule(ResourceMappingProxyGroupGBPTestCase,
                     test_resource_mapping.TestPolicyRule):
    pass


class TestExternalSegment(ResourceMappingProxyGroupGBPTestCase,
                          test_resource_mapping.TestExternalSegment):
    def test_update(self):
        super(TestExternalSegment, self).test_update(
            proxy_ip_pool1='182.169.0.0/16',
            proxy_ip_pool2='172.169.0.0/16')


class TestExternalPolicy(ResourceMappingProxyGroupGBPTestCase,
                         test_resource_mapping.TestExternalPolicy):
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class NodeDriverBase(object):
    """Node Driver Base class for Node Composition Plugin (NCP).

    A Node Driver is the fundamental unit of the NCP service chain plugin.
    It is invoked every time an operation has to be executed on Service Node
    instances (eg. services that are part of a deployed chain)
    which the Node Driver is capable of deploy, destroy and update.
    The Node Driver may expose resource needs to the NCP plugin, that will
    make sure that the NodeDriverContext is enriched with all that's needed by
    the driver.
    """

    @abc.abstractmethod
    def initialize(self, name):
        """Perform driver initialization.

        Called after all drivers have been loaded and the database has
        been initialized. No abstract methods defined below will be
        called prior to this method being called. Name is a unique attribute
        that identifies the driver.
        """
        pass

    @abc.abstractmethod
    def get_plumbing_info(self, context):
        """ Tells the NCP Plugin which kind of plumbing is needed by the Node.

        The plumbing info is defined as a collection of needed policy targets
        on a specific role, this may vary based on the node
        (obtained from the NodeDriverContext) that the specific driver is asked
        to deploy. An example of plumbing info is the following:

        {
            "management": <list of updated PT body dicts, one for each needed>,
            "provider": <list of updated PT body dicts, one for each needed>,
            "consumer": <list of updated PT body dicts, one for each needed>
        }

        The role (key of the above dictionary) specifies in which "side" the
        policy target has to exist. Depending on the kind of chaining the
        Neutron port could actually be placed somewhere else! The value
        is a list of attributes intended to override the PT body. This could
        be used, for example, for providing explicit Neutron Ports when the
        driver requires it or for establishing a naming convention for the PTs.
        An empty dictionary will be mostly used in this case, which will
        indicate a basic PT creation:

        {
            "management": [{}],  # One PT needed in the management
            "provider": [{}, {port_id: 'a'}],  # Two PT needed in the provider
            "consumer": []  # Zero PT needed in the consumer
        }
        """
        pass

    @abc.abstractmethod
    def validate_create(self, context):
        """Validate whether a SCN can be processed or not for creation.

        This method is intended as a indicative measure of whether the NCP
        plugin should use this specific driver for scheduling a given node.
        A successful validation is a prerequisite but doesn't guarantee that
        this driver will ultimately be chosen.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractmethod
    def validate_update(self, context):
        """Validate whether a SCN can be processed or not.

        This method will be called whenever a specific Node owned by this
        driver needs to be updated. It should be used to verify whether the
        Driver is capable of enforcing the update or not.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractmethod
    def create(self, context):
        """Instantiate a Service Chain Node based on the chain context.

        This method will be called at Service Chain instantiation time by the
        NCP plugin. Every scheduled Node Driver will be assigned a Node of the
        chain that has to be deployed based on the node definition and the
        service chain context. The same driver could be called multiple times
        on different nodes of the same chain.
        The datapath is expected to work according to the user intent at the
        end of the chain instantiation.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractmethod
    def delete(self, context):
        """Destroy a deployed Service Chain Node.

        This method will be called when a Service Chain Instance is destroyed
        or in case of node rescheduling. The driver is expected to undeploy the
        specific node and free the owned resources. Freeing the resources
        created by the NCP plugin as a consequence of the plumbing_info
        method belongs to the NCP plugin, and it is in charge of disposing
        them if needed.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractmethod
    def update(self, context):
        """Update a deployed Service Chain Node.

        Some changes in the Service Chain Node could need modifications in all
        its instances. This method will be used in order to synchronize the
        service configuration with the user expectation.
        The original node definition is provided in the context in order to
        calculate the difference if needed.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractmethod
    def update_policy_target_added(self, context, policy_target):
        """Update a deployed Service Chain Node on adding of a PT.

        This method can be used for auto scaling some services whenever a
        Policy Target is added to a relevant PTG.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        :param policy_target: Dict representing a Policy Target.
        """
        pass

    @abc.abstractmethod
    def update_policy_target_removed(self, context, policy_target):
        """Update a deployed Service Chain Node on removal of a PT.

        This method can be used for auto scaling some services whenever a
        Policy Target is removed from a relevant PTG.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        :param policy_target: Dict representing a Policy Target.
        """
        pass

    @abc.abstractmethod
    def update_node_consumer_ptg_added(self, context, policy_target_group):
        """Update a deployed Service Chain Node on addition of a consumer PTG.

        This method can be used for auto scaling some services whenever a
        Policy Target is added to a relevant PTG.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        :param policy_target_group: Dict representing a Policy Target Group.
        """
        pass

    @abc.abstractmethod
    def update_node_consumer_ptg_removed(self, context, policy_target_group):
        """Update a deployed Service Chain Node on removal of a consumer PTG.

        This method can be used for auto scaling some services whenever a
        Policy Target is removed from a relevant PTG.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        :param policy_target_group: Dict representing a Policy Target Group.
        """
        pass

    @abc.abstractmethod
    def policy_target_group_updated(self, context, old_policy_target_group,
                                current_policy_target_group):
        """Update a Node Driver that a chain provider PTG was created or
        changed
        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        :param old_policy_target_group: PTG state before the update. None if
        the group was created.
        :param current_policy_target_group: Current PTG state.
        """
        pass

    @abc.abstractmethod
    def notify_chain_parameters_updated(self, context):
        """Update a deployed Service Chain Node on GBP PRS updates

        This method can be used to inform the node driver that some parameter
        that affects the service chain is updated. The update may be
        something like adding or removing an Allow Rule to the ruleset and
        this has to be enforced in the Firewall Service VM, or it could simply
        be a classifier update.

        :param context: NodeDriverContext instance describing the service chain
        and the specific node to be processed by this driver.
        """
        pass

    @abc.abstractproperty
    def name(self):
        pass



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from neutron.api import extensions
from neutron.conf.plugins.ml2 import config  # noqa
from neutron.conf.plugins.ml2.drivers import driver_type
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit.extensions import test_address_scope
from neutron_lib.agent import topics
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_config import cfg

from gbpservice.neutron.db import all_models  # noqa
import gbpservice.neutron.extensions
from gbpservice.neutron.tests.unit.plugins.ml2plus.drivers import (
    mechanism_logger as mech_logger)

PLUGIN_NAME = 'gbpservice.neutron.plugins.ml2plus.plugin.Ml2PlusPlugin'

config.register_ml2_plugin_opts()
driver_type.register_ml2_drivers_vlan_opts()


# This is just a quick sanity test that basic ML2 plugin functionality
# is preserved.

class Ml2PlusPluginV2TestCase(test_address_scope.AddressScopeTestCase):

    def setUp(self):
        # Enable the test mechanism driver to ensure that
        # we can successfully call through to all mechanism
        # driver apis.
        cfg.CONF.set_override('mechanism_drivers',
                ['logger_plus', 'test'], group='ml2')
        cfg.CONF.set_override('network_vlan_ranges',
                ['physnet1:1000:1099'], group='ml2_type_vlan')

        extensions.append_api_extensions_path(
            gbpservice.neutron.extensions.__path__)
        super(Ml2PlusPluginV2TestCase, self).setUp(PLUGIN_NAME)
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.port_create_status = 'DOWN'
        self.plugin = directory.get_plugin()

    def exist_checker(self, getter):
        def verify(context):
            obj = getter(context._plugin_context, context.current['id'])
            self.assertIsNotNone(obj)
            return mock.DEFAULT
        return verify


class TestRpcListeners(Ml2PlusPluginV2TestCase):
    @staticmethod
    def _consume_in_threads(self):
        return self.servers

    @staticmethod
    def _start_rpc_listeners(self):
        conn = n_rpc.Connection()
        conn.create_consumer('q-test-topic', [])
        return conn.consume_in_threads()

    def test_start_rpc_listeners(self):
        # Override mock from neutron_lib.fixture.RPCFixture installed
        # by neutron.tests.base.BaseTestCase.setUp(), so that it
        # returns servers, but still avoids starting them.
        with mock.patch.object(
                n_rpc.Connection, 'consume_in_threads',
                TestRpcListeners._consume_in_threads):
            # Mock logger MD to start an RPC listener.
            with mock.patch(
                    'gbpservice.neutron.tests.unit.plugins.ml2plus.drivers.'
                    'mechanism_logger.LoggerPlusMechanismDriver.'
                    'start_rpc_listeners',
                    TestRpcListeners._start_rpc_listeners):
                # Call plugin method and verify that the base ML2
                # servers as well as the test MD server are returned.
                servers = self.plugin.start_rpc_listeners()
                self.assertEqual(sorted([topics.PLUGIN,
                                         topics.SERVER_RESOURCE_VERSIONS,
                                         topics.REPORTS,
                                         'q-test-topic']),
                                 sorted([server._target.topic
                                         for server in servers]))


class TestEnsureTenant(Ml2PlusPluginV2TestCase):
    def test_network(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            self._make_network(self.fmt, 'net', True, tenant_id='t1')
            et.assert_has_calls([mock.call(mock.ANY, 't1')])
            self.assertEqual(2, et.call_count)

    def test_network_bulk(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            networks = [{'network': {'name': 'n1',
                                     'tenant_id': 't1'}},
                        {'network': {'name': 'n2',
                                     'tenant_id': 't2'}},
                        {'network': {'name': 'n3',
                                     'tenant_id': 't1'}}]
            res = self._create_bulk_from_list(self.fmt, 'network', networks)
            self.assertEqual(201, res.status_int)
            et.assert_has_calls([mock.call(mock.ANY, 't1'),
                                 mock.call(mock.ANY, 't2')],
                                any_order=True)
            self.assertEqual(4, et.call_count)

    def test_subnet(self):
        net = self._make_network(self.fmt, 'net', True)

        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            self._make_subnet(self.fmt, net, None, '10.0.0.0/24',
                              tenant_id='t1')
            et.assert_called_once_with(mock.ANY, 't1')

    def test_subnet_bulk(self):
        net = self._make_network(self.fmt, 'net', True)
        network_id = net['network']['id']

        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            subnets = [{'subnet': {'name': 's1',
                                   'network_id': network_id,
                                   'ip_version': 4,
                                   'cidr': '10.0.1.0/24',
                                   'tenant_id': 't1'}},
                       {'subnet': {'name': 's2',
                                   'network_id': network_id,
                                   'ip_version': 4,
                                   'cidr': '10.0.2.0/24',
                                   'tenant_id': 't2'}},
                       {'subnet': {'name': 'n3',
                                   'network_id': network_id,
                                   'ip_version': 4,
                                   'cidr': '10.0.3.0/24',
                                   'tenant_id': 't1'}}]
            res = self._create_bulk_from_list(self.fmt, 'subnet', subnets)
            self.assertEqual(201, res.status_int)
            et.assert_has_calls([mock.call(mock.ANY, 't1'),
                                 mock.call(mock.ANY, 't2')],
                                any_order=True)
            self.assertEqual(2, et.call_count)

    def test_port(self):
        net = self._make_network(self.fmt, 'net', True)

        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            self._make_port(self.fmt, net['network']['id'], tenant_id='t1')
            et.assert_has_calls([mock.call(mock.ANY, 't1')])
            self.assertEqual(2, et.call_count)

    def test_port_bulk(self):
        net = self._make_network(self.fmt, 'net', True)
        network_id = net['network']['id']

        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            ports = [{'port': {'name': 's1',
                               'network_id': network_id,
                               'tenant_id': 't1'}},
                     {'port': {'name': 's2',
                               'network_id': network_id,
                               'tenant_id': 't2'}},
                     {'port': {'name': 'n3',
                               'network_id': network_id,
                               'tenant_id': 't1'}}]
            res = self._create_bulk_from_list(self.fmt, 'port', ports)
            self.assertEqual(201, res.status_int)
            et.assert_has_calls([mock.call(mock.ANY, 't1'),
                                 mock.call(mock.ANY, 't2')],
                                any_order=True)
            self.assertEqual(4, et.call_count)

    def test_subnetpool(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            self._make_subnetpool(self.fmt, ['10.0.0.0/8'], name='sp1',
                                  tenant_id='t1')
            et.assert_called_once_with(mock.ANY, 't1')

    def test_address_scope(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'ensure_tenant') as et:
            self._make_address_scope(self.fmt, 4, name='as1', tenant_id='t1')
            et.assert_called_once_with(mock.ANY, 't1')


class TestSubnetPool(Ml2PlusPluginV2TestCase):
    def test_create(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'create_subnetpool_precommit') as pre:
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'create_subnetpool_postcommit') as post:
                self._make_subnetpool(self.fmt, ['10.0.0.0/8'], name='sp1',
                                      tenant_id='t1')

                self.assertEqual(1, pre.call_count)
                self.assertEqual('sp1',
                                 pre.call_args[0][0].current['name'])
                self.assertIsNone(pre.call_args[0][0].original)

                self.assertEqual(1, post.call_count)
                self.assertEqual('sp1',
                                 post.call_args[0][0].current['name'])
                self.assertIsNone(post.call_args[0][0].original)

    def test_update(self):
        subnetpool = self._make_subnetpool(
            self.fmt, ['10.0.0.0/8'], name='sp1', tenant_id='t1')['subnetpool']
        data = {'subnetpool': {'name': 'newnameforsubnetpool'}}
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'update_subnetpool_precommit') as pre:
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'update_subnetpool_postcommit') as post:
                res = self._update('subnetpools', subnetpool['id'],
                                   data)['subnetpool']
                self.assertEqual('newnameforsubnetpool', res['name'])

                self.assertEqual(1, pre.call_count)
                self.assertEqual('newnameforsubnetpool',
                                 pre.call_args[0][0].current['name'])
                self.assertEqual('sp1',
                                 pre.call_args[0][0].original['name'])

                self.assertEqual(1, post.call_count)
                self.assertEqual('newnameforsubnetpool',
                                 post.call_args[0][0].current['name'])
                self.assertEqual('sp1',
                                 post.call_args[0][0].original['name'])

    def test_delete(self):
        subnetpool = self._make_subnetpool(
            self.fmt, ['10.0.0.0/8'], name='sp1', tenant_id='t1')['subnetpool']
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'delete_subnetpool_precommit') as pre:
            pre.side_effect = self.exist_checker(
                                self.plugin.get_subnetpool)
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'delete_subnetpool_postcommit') as post:
                self._delete('subnetpools', subnetpool['id'])

                self.assertEqual(1, pre.call_count)
                self.assertEqual('sp1',
                                 pre.call_args[0][0].current['name'])
                self.assertIsNone(pre.call_args[0][0].original)

                self.assertEqual(1, post.call_count)
                self.assertEqual('sp1',
                                 post.call_args[0][0].current['name'])
                self.assertIsNone(post.call_args[0][0].original)


class TestAddressScope(Ml2PlusPluginV2TestCase):
    def test_create(self):
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'create_address_scope_precommit') as pre:
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'create_address_scope_postcommit') as post:
                self._make_address_scope(self.fmt, 4, name='as1',
                                         tenant_id='t1')

                self.assertEqual(1, pre.call_count)
                self.assertEqual('as1',
                                 pre.call_args[0][0].current['name'])
                self.assertIsNone(pre.call_args[0][0].original)

                self.assertEqual(1, post.call_count)
                self.assertEqual('as1',
                                 post.call_args[0][0].current['name'])
                self.assertIsNone(post.call_args[0][0].original)

    def test_update(self):
        address_scope = self._make_address_scope(
            self.fmt, 4, name='as1', tenant_id='t1')['address_scope']
        data = {'address_scope': {'name': 'newnameforaddress_scope'}}
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'update_address_scope_precommit') as pre:
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'update_address_scope_postcommit') as post:
                res = self._update('address-scopes', address_scope['id'],
                                   data)['address_scope']
                self.assertEqual('newnameforaddress_scope', res['name'])

                self.assertEqual(1, pre.call_count)
                self.assertEqual('newnameforaddress_scope',
                                 pre.call_args[0][0].current['name'])
                self.assertEqual('as1',
                                 pre.call_args[0][0].original['name'])

                self.assertEqual(1, post.call_count)
                self.assertEqual('newnameforaddress_scope',
                                 post.call_args[0][0].current['name'])
                self.assertEqual('as1',
                                 post.call_args[0][0].original['name'])

    def test_delete(self):
        address_scope = self._make_address_scope(
            self.fmt, 4, name='as1', tenant_id='t1')['address_scope']
        with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                               'delete_address_scope_precommit') as pre:
            pre.side_effect = self.exist_checker(self.plugin.get_address_scope)
            with mock.patch.object(mech_logger.LoggerPlusMechanismDriver,
                                   'delete_address_scope_postcommit') as post:
                self._delete('address-scopes', address_scope['id'])

                self.assertEqual(1, pre.call_count)
                self.assertEqual('as1',
                                 pre.call_args[0][0].current['name'])
                self.assertIsNone(pre.call_args[0][0].original)

                self.assertEqual(1, post.call_count)
                self.assertEqual('as1',
                                 post.call_args[0][0].current['name'])
                self.assertIsNone(post.call_args[0][0].original)


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2BasicGet(test_plugin.TestBasicGet,
                      Ml2PlusPluginV2TestCase):
    pass


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2V2HTTPResponse(test_plugin.TestV2HTTPResponse,
                            Ml2PlusPluginV2TestCase):
    pass


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2PortsV2(test_plugin.TestPortsV2,
                     Ml2PlusPluginV2TestCase):
    pass


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2NetworksV2(test_plugin.TestNetworksV2,
                        Ml2PlusPluginV2TestCase):
    pass


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2SubnetsV2(test_plugin.TestSubnetsV2,
                       Ml2PlusPluginV2TestCase):
    pass


# REVISIT: Skipping inherited ML2 tests to reduce UT run time.
@testtools.skip('Skipping test class')
class TestMl2SubnetPoolsV2(test_plugin.TestSubnetPoolsV2,
                           Ml2PlusPluginV2TestCase):
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""ip_pool & subnet_prefix_length nullable

Revision ID: 4af01d620224
Revises: 75aa8a37a8de
Create Date: 2016-11-27 21:33:13.384981

"""

# revision identifiers, used by Alembic.
revision = '4af01d620224'
down_revision = '75aa8a37a8de'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('gp_l3_policies', 'ip_pool', nullable=True,
                    existing_nullable=False,
                    existing_type=sa.String(length=64))
    op.alter_column('gp_l3_policies', 'subnet_prefix_length', nullable=True,
                    existing_nullable=False, existing_type=sa.Integer)


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources

from vmware_nsx.db import db as nsx_db
from vmware_nsx.plugins.nsx_v3 import utils as nsx_utils

from vmware_nsxlib import v3
from vmware_nsxlib.v3 import config
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import resources as nsx_resources

from gbpservice._i18n import _
from gbpservice.neutron.services.grouppolicy.common import constants as g_const
from gbpservice.neutron.services.grouppolicy.common import exceptions as gpexc
from gbpservice.neutron.services.grouppolicy.drivers import (
    resource_mapping as api)


LOG = logging.getLogger(__name__)

SINGLE_ENTRY_ID = 'GBP'
DRIVER_NAME = 'NSX Policy driver'
DRIVER_OPT_GROUP = 'NSX_POLICY'
NSX_V3_GROUP = 'nsx_v3'

policy_opts = [
    cfg.StrOpt('nsx_policy_manager',
               help=_("Nsx Policy manager IP address or host.")),
    cfg.StrOpt('nsx_policy_username',
               help=_("Nsx Policy username.")),
    cfg.StrOpt('nsx_policy_password',
               help=_("Nsx Policy password.")),
    cfg.StrOpt('nsx_manager_thumbprint',
               help=_("Thumbprint of nsx manager"))
]

cfg.CONF.register_opts(policy_opts, DRIVER_OPT_GROUP)


class HierarchicalContractsNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Hierarchy in rule sets is not supported with %s." %
               DRIVER_NAME)


class UpdateOperationNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Update operation on this object is not supported with %s." %
               DRIVER_NAME)


class ProxyGroupsNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Proxy groups are not supported with %s." % DRIVER_NAME)


# TODO(annak): remove when ipv6 is supported + add support for ICMPv6 service
class Ipv6NotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Ipv6 not supported with %s" % DRIVER_NAME)


class UpdateClassifierProtocolNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Update operation on classifier protocol is not supported "
               "with %s" % DRIVER_NAME)


class UpdateClassifierDirectionNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Update operation on classifier direction is not supported "
               "with %s" % DRIVER_NAME)


class ProtocolNotSupported(gpexc.GroupPolicyBadRequest):
    message = ("Unsupported classifier protocol. Only icmp, tcp and udp are "
               "supported with %s" % DRIVER_NAME)


def append_in_dir(name):
    return name + '_I'


def append_out_dir(name):
    return name + '_O'


def generate_nsx_name(uuid, name, tag=None, maxlen=80):
    short_uuid = '_' + uuid[:5] + '...' + uuid[-5:]
    maxlen = maxlen - len(short_uuid)
    if not name:
        name = ''
    if tag:
        maxlen = maxlen - len(tag) - 1
        return name[:maxlen] + '_' + tag + short_uuid
    else:
        return name[:maxlen] + short_uuid


class NsxPolicyMappingDriver(api.ResourceMappingDriver):
    """Nsx Policy Mapping driver for Group Policy plugin.

    This mapping driver is only supported with nsxv3 core plugin.
    NSX Manager is the network virtualization appliance configured by the core
    plugin.
    NSX Policy is a separate appliance that provides grouping API. Behind the
    scenes, NSX Policy configures same NSX manager.

    At current phase of development, security is achieved via NSX Policy,
    while connectivity functionality is inherited from resource mapping driver.

    This driver configures services, connectivity rules and grouping objects
    on NSX Policy. In addition, it configures logical port tag directly on
    NSX manager, in order to provide port membership in the desired group.

    The driver does not maintain state of its own (no db extension). This is
    for sake of reducing failure recovery problems, at cost of making few more
    backend roundtrips.
    """
    def get_nsxpolicy_lib(self):
        """ Prepare agent for NSX Policy API calls"""
        nsxlib_config = config.NsxLibConfig(
            nsx_api_managers=[cfg.CONF.NSX_POLICY.nsx_policy_manager],
            username=cfg.CONF.NSX_POLICY.nsx_policy_username,
            password=cfg.CONF.NSX_POLICY.nsx_policy_password)

        return v3.NsxPolicyLib(nsxlib_config)

    def get_nsxmanager_lib(self):
        """Prepare agent for NSX Manager API calls"""
        return nsx_utils.get_nsxlib_wrapper()

    def initialize(self):
        super(NsxPolicyMappingDriver, self).initialize()
        self._gbp_plugin = None
        self.nsx_policy = self.get_nsxpolicy_lib()
        # reinitialize the cluster upon fork for api workers to ensure each
        # process has its own keepalive loops + state
        registry.subscribe(
            self.nsx_policy.reinitialize_cluster,
            resources.PROCESS, events.AFTER_INIT)
        self.policy_api = self.nsx_policy.policy_api

        self.nsx_manager = self.get_nsxmanager_lib()
        registry.subscribe(
            self.nsx_manager.reinitialize_cluster,
            resources.PROCESS, events.AFTER_INIT)

        self.nsx_port = nsx_resources.LogicalPort(self.nsx_manager.client)

        self._verify_enforcement_point()

        # TODO(annak): add validation for core plugin (can only be nsxv3)

    def _verify_enforcement_point(self):
        """Configure NSX Policy to enforce grouping rules on NSX Manager"""

        # We only support a single NSX manager at this point
        nsx_manager_ip = cfg.CONF.nsx_v3.nsx_api_managers[0]
        nsx_manager_username = cfg.CONF.nsx_v3.nsx_api_user[0]
        nsx_manager_password = cfg.CONF.nsx_v3.nsx_api_password[0]
        nsx_manager_thumbprint = cfg.CONF.NSX_POLICY.nsx_manager_thumbprint
        epoints = self.nsx_policy.enforcement_point.list()
        for ep in epoints:
            conn = ep['connection_info']
            if conn and conn['enforcement_point_address'] == nsx_manager_ip:
                LOG.debug('Enforcement point for %s already exists (%s)',
                          nsx_manager_ip, ep['id'])
                return

        LOG.info('Creating enforcement point for %s', nsx_manager_ip)
        self.nsx_policy.enforcement_point.create_or_overwrite(
            name=nsx_manager_ip,
            ep_id=SINGLE_ENTRY_ID,
            ip_address=nsx_manager_ip,
            username=nsx_manager_username,
            password=nsx_manager_password,
            thumbprint=nsx_manager_thumbprint)

    def _create_domain(self, context):
        project_id = context.current['project_id']
        tenant_name = context._plugin_context.tenant_name
        domain_name = generate_nsx_name(project_id, tenant_name)

        LOG.info('Creating domain %(domain)s for project %(project)s',
                 {'domain': domain_name,
                  'project': project_id})

        self.nsx_policy.domain.create_or_overwrite(
            name=domain_name,
            domain_id=project_id,
            description=_('Domain for tenant %s') % tenant_name)

        self.nsx_policy.deployment_map.create_or_overwrite(
            name=domain_name,
            map_id=project_id,
            domain_id=project_id,
            ep_id=SINGLE_ENTRY_ID)

    def _delete_domain(self, project_id):
        try:
            self.nsx_policy.deployment_map.delete(project_id,
                                                  domain_id=project_id)
        except nsxlib_exc.ManagerError:
            LOG.warning('Domain %s is not deployed on backend',
                        project_id)

        try:
            self.nsx_policy.domain.delete(project_id)
        except nsxlib_exc.ManagerError:
            LOG.warning('Domain %s was not found on backend',
                        project_id)

    def _split_rules_by_direction(self, context, rules):
        in_dir = [g_const.GP_DIRECTION_BI, g_const.GP_DIRECTION_IN]
        out_dir = [g_const.GP_DIRECTION_BI, g_const.GP_DIRECTION_OUT]

        in_rules = []
        out_rules = []

        for rule in rules:
            classifier = context._plugin.get_policy_classifier(
                                             context._plugin_context,
                                             rule['policy_classifier_id'])
            direction = classifier['direction']
            if direction in in_dir:
                in_rules.append(rule)

            if direction in out_dir:
                out_rules.append(rule)

        return in_rules, out_rules

    def _get_services_from_rule_set(self, context, rule_set_id):

        ruleset = self.gbp_plugin.get_policy_rule_set(
                context._plugin_context, rule_set_id)
        rules = self.gbp_plugin.get_policy_rules(
                context._plugin_context,
                {'id': ruleset['policy_rules']})

        in_rules, out_rules = self._split_rules_by_direction(context, rules)
        in_services = set()
        for rule in in_rules:
            in_services.add(rule['policy_classifier_id'])
        out_services = set()
        for rule in out_rules:
            out_services.add(rule['policy_classifier_id'])

        return sorted(list(in_services)), sorted(list(out_services))

    def _filter_ptgs_by_ruleset(self, ptgs, ruleset_id):
        providing_ptgs = [ptg['id'] for ptg in ptgs
                          if ruleset_id in ptg['provided_policy_rule_sets']]
        consuming_ptgs = [ptg['id'] for ptg in ptgs
                          if ruleset_id in ptg['consumed_policy_rule_sets']]
        return providing_ptgs, consuming_ptgs

    def _map_rule_set(self, context, ptgs, project_id,
                      ruleset_id, delete_flow):

        def delete_map_if_exists(ruleset):
            try:
                self.nsx_policy.comm_map.delete(project_id, ruleset)
            except nsxlib_exc.ManagerError:
                # TODO(annak) - narrow this exception down
                pass

        providing_ptgs, consuming_ptgs = self._filter_ptgs_by_ruleset(
            ptgs, ruleset_id)

        ruleset_in = append_in_dir(ruleset_id)
        ruleset_out = append_out_dir(ruleset_id)
        if not consuming_ptgs or not providing_ptgs:
            if not delete_flow:
                return
            if not consuming_ptgs and not providing_ptgs:
                return

            # we need to delete map entry if exists
            for ruleset in (ruleset_in, ruleset_out):
                delete_map_if_exists(ruleset)

            return

        services_in, services_out = self._get_services_from_rule_set(
            context, ruleset_id)

        if services_in:
            self.nsx_policy.comm_map.create_or_overwrite(
                name=ruleset_in,
                domain_id=project_id,
                map_id=ruleset_in,
                description="GBP ruleset ingress",
                service_ids=services_in,
                source_groups=consuming_ptgs,
                dest_groups=providing_ptgs)
        else:
            delete_map_if_exists(ruleset_in)

        if services_out:
            self.nsx_policy.comm_map.create_or_overwrite(
                name=ruleset_out,
                domain_id=project_id,
                map_id=ruleset_out,
                description="GBP ruleset egress",
                service_ids=services_out,
                source_groups=providing_ptgs,
                dest_groups=consuming_ptgs)
        else:
            delete_map_if_exists(ruleset_out)

    def _map_group_rule_sets(self, context,
                             provided_policy_rule_sets,
                             consumed_policy_rule_sets,
                             delete_flow=False):

        project_id = context.current['project_id']

        # create communication maps
        ptgs = context._plugin.get_policy_target_groups(
            context._plugin_context)
        for ruleset in provided_policy_rule_sets:
            self._map_rule_set(context, ptgs, project_id,
                               ruleset, delete_flow)

        for ruleset in consumed_policy_rule_sets:
            self._map_rule_set(context, ptgs, project_id,
                               ruleset, delete_flow)

    # overrides base class, called from base group_create_postcommit
    # REVISIT(annak): Suggest a better design for driver-specific callbacks,
    # based on connectivity vs. security
    def _set_sg_rules_for_subnets(self, context, subnets,
                                  provided_policy_rule_sets,
                                  consumed_policy_rule_sets):
        pass

    # overrides base class, called from base group_delete_postcommit
    def _unset_sg_rules_for_subnets(self, context, subnets,
                                    provided_policy_rule_sets,
                                    consumed_policy_rule_sets):
        pass

    # Overrides base class
    def _update_sgs_on_ptg(self, context, ptg_id,
                           provided_policy_rule_sets,
                           consumed_policy_rule_sets, op):

        self._map_group_rule_sets(
             context,
             provided_policy_rule_sets,
             consumed_policy_rule_sets,
             delete_flow=(op == "DISASSOCIATE"))

    def create_policy_action_precommit(self, context):
        pass

    def create_policy_action_postcommit(self, context):
        super(NsxPolicyMappingDriver,
              self).create_policy_action_postcommit(context)

    def create_policy_classifier_precommit(self, context):
        if context.current['protocol'] not in ('icmp', 'tcp', 'udp'):
            raise ProtocolNotSupported()

    def create_policy_classifier_postcommit(self, context):
        classifier = context.current

        if classifier['protocol'] == 'icmp':
            self.nsx_policy.icmp_service.create_or_overwrite(
                name=classifier['name'],
                service_id=classifier['id'],
                description=classifier['description'])
            return

        ports = []
        if classifier['port_range']:
            port_range = classifier['port_range'].split(':', 1)
            lower = int(port_range[0])
            upper = int(port_range[-1]) + 1
            ports = [str(p) for p in range(lower, upper)]

        # service entry in nsx policy has single direction
        # directions will be enforced on communication map level
        self.nsx_policy.service.create_or_overwrite(
            name=generate_nsx_name(classifier['id'], classifier['name']),
            service_id=classifier['id'],
            description=classifier['description'],
            protocol=classifier['protocol'],
            dest_ports=ports)

    def create_policy_rule_precommit(self, context):
        pass

    def create_policy_rule_postcommit(self, context, transaction=None):
        pass

    def create_policy_rule_set_precommit(self, context):
        if context.current['child_policy_rule_sets']:
            raise HierarchicalContractsNotSupported()

    def create_policy_rule_set_postcommit(self, context):
        pass

    def create_policy_target_precommit(self, context):
        super(NsxPolicyMappingDriver,
              self).create_policy_target_precommit(context)

    def _tag_port(self, context, port_id, tag):
        # Translate neutron port id to nsx port id
        _net_id, nsx_port_id = nsx_db.get_nsx_switch_and_port_id(
                 context._plugin_context.session, port_id)

        self.nsx_port.update(nsx_port_id, None,
                             tags_update=[{'scope': 'gbp',
                                           'tag': tag}])

    def _get_project_ptgs(self, context, project_id):
        ptgs = context._plugin.get_policy_target_groups(
                context._plugin_context)

        return [ptg for ptg in ptgs if ptg['project_id'] == project_id]

    def create_policy_target_postcommit(self, context):
        if not context.current['port_id']:
            self._use_implicit_port(context)
        self._tag_port(context,
                       context.current['port_id'],
                       context.current['policy_target_group_id'])

        # Below is inherited behaviour
        self._update_cluster_membership(
                context, new_cluster_id=context.current['cluster_id'])
        self._associate_fip_to_pt(context)

    def create_policy_target_group_precommit(self, context):
        if context.current.get('proxied_group_id'):
            raise ProxyGroupsNotSupported()

        super(NsxPolicyMappingDriver,
              self).create_policy_target_group_precommit(context)

    def create_policy_target_group_postcommit(self, context):
        # create the group on backend
        group_id = context.current['id']
        project_id = context.current['project_id']

        # create the domain for this project if needed
        project_ptgs = self._get_project_ptgs(context, project_id)
        if len(project_ptgs) == 1:
            # we've just created the first group for this project
            # need to create a domain for the project on backend
            self._create_domain(context)

        self.nsx_policy.group.create_or_overwrite(
            name=generate_nsx_name(group_id, context.current['name']),
            domain_id=project_id,
            group_id=group_id,
            description=context.current['description'],
            cond_val=group_id)

        # This will take care of connectivity and invoke overriden
        # callbacks defined above for security
        super(NsxPolicyMappingDriver,
              self).create_policy_target_group_postcommit(context)

    def delete_policy_target_group_postcommit(self, context):
        group_id = context.current['id']
        project_id = context.current['project_id']
        self.nsx_policy.group.delete(project_id, group_id)

        # delete the domain for this project if needed
        project_ptgs = self._get_project_ptgs(context, project_id)
        if len(project_ptgs) == 0:
            # we've just deleted the last group for this project
            # need to clean up the project domain on backend
            self._delete_domain(project_id)

        # This will take care of connectivity and invoke overriden
        # callbacks defined above for security
        super(NsxPolicyMappingDriver,
              self).delete_policy_target_group_postcommit(context)

    def delete_policy_classifier_precommit(self, context):
        pass

    def delete_policy_classifier_postcommit(self, context):
        classifier_id = context.current['id']
        if context.current['protocol'] == 'icmp':
            self.nsx_policy.icmp_service.delete(classifier_id)
        else:
            self.nsx_policy.service.delete(classifier_id)

    def delete_policy_rule_set_precommit(self, context):
        pass

    def delete_policy_rule_set_postcommit(self, context):
        pass

    def delete_policy_target_postcommit(self, context):
        # This is inherited behavior without:
        # 1. sg disassociation
        # 2. proxy handling
        port_id = context.current['port_id']
        for fip in context.fips:
            self._delete_fip(context._plugin_context,
                             fip.floatingip_id)
        self._cleanup_port(context._plugin_context, port_id)

    def update_policy_rule_set_precommit(self, context):
        self._reject_shared(context.current, 'policy_rule_set')

    def update_policy_rule_set_postcommit(self, context):
        if (context.current['policy_rules'] !=
            context.original['policy_rules']):
            self._on_policy_rule_set_updated(context, context.current)

    def update_policy_target_precommit(self, context):
        # Parent call verifies change of PTG is not supported
        super(NsxPolicyMappingDriver,
              self).update_policy_target_precommit(context)

    def update_policy_target_postcommit(self, context):
        # Since change of PTG is not supported, nothing to add here
        super(NsxPolicyMappingDriver,
              self).update_policy_target_postcommit(context)

    def update_policy_rule_precommit(self, context):
        super(NsxPolicyMappingDriver,
              self).update_policy_rule_precommit(context)

    def _on_policy_rule_set_updated(self, context, prs):
        ptgs = context._plugin.get_policy_target_groups(
            context._plugin_context)

        self._map_rule_set(context, ptgs, prs['project_id'],
                           prs['id'], False)

    def update_policy_rule_postcommit(self, context):
        if (context.current['policy_classifier_id'] !=
            context.original['policy_classifier_id']):
            # All groups using this rule need to be updated
            prs_ids = (
                context._plugin._get_policy_rule_policy_rule_sets(
                    context._plugin_context, context.current['id']))
            policy_rule_sets = context._plugin.get_policy_rule_sets(
                context._plugin_context, filters={'id': prs_ids})
            for prs in policy_rule_sets:
                self._on_policy_rule_set_updated(context, prs)

    def update_policy_action_precommit(self, context):
        raise UpdateOperationNotSupported()

    def update_policy_classifier_precommit(self, context):
        if context.current['protocol'] != context.original['protocol']:
            raise UpdateClassifierProtocolNotSupported()

        # TODO(annak): support this
        if context.current['direction'] != context.original['direction']:
            raise UpdateClassifierDirectionNotSupported()

    def update_policy_classifier_postcommit(self, context):
        self.create_policy_classifier_postcommit(context)

    def create_l3_policy_precommit(self, context):
        if context.current['ip_version'] != 4:
            raise Ipv6NotSupported()

        super(NsxPolicyMappingDriver,
              self).create_l3_policy_precommit(context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Table for cisco_apic network extra contracts extension attributes

Revision ID: f28141ea1bbf
Revises: b76dc22f2e23
Create Date: 2019-12-20 13:08:03.603312

"""

# revision identifiers, used by Alembic.
revision = 'f28141ea1bbf'
down_revision = 'b76dc22f2e23'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_network_extra_contracts',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('contract_name', sa.String(64), nullable=False),
        sa.Column('provides', sa.Boolean, nullable=False),
        sa.ForeignKeyConstraint(
            ['network_id'], ['networks.id'],
            name='apic_aim_network_contract_extn_fk_network',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'contract_name', 'provides')
    )


def downgrade():
    pass



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""sc_node config datatype change

Revision ID: 31b399f08b1c
Revises: 7ef98f287d6
Create Date: 2015-12-22 09:15:44.917899

"""

# revision identifiers, used by Alembic.
revision = '31b399f08b1c'
down_revision = '7ef98f287d6'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('sc_nodes', 'config',
                    existing_type=sa.String(4096),
                    type_=sa.TEXT)


def downgrade():
    pass



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

""" gbp_db_network_service_policy

Revision ID: 577bb4469944
"""

# revision identifiers, used by Alembic.
revision = '577bb4469944'
down_revision = '6d76bcf836a7'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_network_service_policies',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_network_service_params',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('param_type', sa.String(length=50), nullable=False),
        sa.Column('param_name', sa.String(length=50), nullable=False),
        sa.Column('param_value', sa.String(length=50), nullable=False),
        sa.Column('network_service_policy_id', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['network_service_policy_id'],
                                ['gp_network_service_policies.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'))

    op.add_column(
        'gp_policy_target_groups',
        sa.Column('network_service_policy_id',
                  sa.String(length=36), nullable=True))

    op.create_unique_constraint(None, 'gp_policy_target_groups',
                                ['network_service_policy_id'])

    op.create_foreign_key('gp_policy_target_groups_ibfk_nsp',
                          source='gp_policy_target_groups',
                          referent='gp_network_service_policies',
                          local_cols=['network_service_policy_id'],
                          remote_cols=['id'], ondelete='CASCADE')


def downgrade():

    op.drop_constraint('gp_policy_target_groups_ibfk_nsp',
                       'gp_policy_target_groups',
                       'foreignkey')
    op.drop_column('gp_policy_target_groups', 'network_service_policy_id')
    op.drop_table('gp_network_service_params')
    op.drop_table('gp_network_service_policies')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import OrderedDict
import sys

from neutron_lib import context as n_context
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import importutils
from stevedore import driver

from gbpservice._i18n import _

LOG = logging.getLogger(__name__)
cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')


def get_function_local_from_stack(function, local):
    frame = sys._getframe()
    while frame:
        if frame.f_code.co_name == function:
            return frame.f_locals.get(local)
        frame = frame.f_back


def get_obj_from_stack(cls):
    i = 1
    try:
        while True:
            for val in sys._getframe(i).f_locals.values():
                if isinstance(val, cls):
                    return val
            i = i + 1
    except Exception:
        return


def get_current_context():
    return get_obj_from_stack(n_context.Context)


def get_resource_plural(resource):
    if resource.endswith('y'):
        resource_plural = resource.replace('y', 'ies')
    else:
        resource_plural = resource + 's'

    return resource_plural


def load_plugin(namespace, plugin):
    try:
        # Try to resolve plugin by name
        mgr = driver.DriverManager(namespace, plugin)
        plugin_class = mgr.driver
    except RuntimeError as e1:
        # fallback to class name
        try:
            plugin_class = importutils.import_class(plugin)
        except ImportError as e2:
            LOG.exception("Error loading plugin by name, %s", e1)
            LOG.exception("Error loading plugin by class, %s", e2)
            raise ImportError(_("Plugin not found."))
    return plugin_class()


def admin_context(context):
    admin_context = n_context.get_admin_context()
    admin_context._session = context.session
    return admin_context


# created a common function for tackling issues of sorting a list of dict
# or list of string for Py3 support
def deep_sort(obj):
    if isinstance(obj, dict):
        obj = OrderedDict(sorted(obj.items()))
        for k, v in obj.items():
            if isinstance(v, dict) or isinstance(v, list):
                obj[k] = deep_sort(v)

    if isinstance(obj, list):
        for i, v in enumerate(obj):
            if isinstance(v, dict) or isinstance(v, list):
                obj[i] = deep_sort(v)
        obj = sorted(
            obj,
            key=lambda x: jsonutils.dumps(x, default=lambda o: o.__dict__))

    return obj


# comparing 2 (nested) dict or list of (nested) dict
def is_equal(obj1, obj2):
    sorted_obj1 = deep_sort(obj1)
    sorted_obj2 = deep_sort(obj2)
    return sorted_obj1 == sorted_obj2


class DictClass(dict):

    def __getattr__(self, item):
        return self[item]

    __setattr__ = dict.__setattr__
    __delattr__ = dict.__delattr__


def get_keystone_creds():
    keystone_conf = cfg.CONF.keystone_authtoken
    user = keystone_conf.username
    pw = keystone_conf.password
    tenant = keystone_conf.project_name
    if keystone_conf.get('auth_uri'):
        auth_url = keystone_conf.auth_uri.rstrip('/')
        if not auth_url.endswith('/v2.0'):
            auth_url += '/v2.0'
    else:
        auth_url = ('%s://%s:%s/v2.0' % (
            keystone_conf.auth_protocol,
            keystone_conf.auth_host,
            keystone_conf.auth_port))
    return user, pw, tenant, auth_url + '/'


def set_difference(iterable_1, iterable_2):
    set1 = set(iterable_1)
    set2 = set(iterable_2)
    return (set1 - set2), (set2 - set1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_segmentation_label_db as db)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_extension_driver_api as test_ext_base)


class ExtensionDriverTestCaseMixin(object):

    def test_pt_lifecycle(self):
        ptg = self.create_policy_target_group()['policy_target_group']

        pt = self.create_policy_target(
            policy_target_group_id=ptg['id'])['policy_target']
        self.assertEqual([], pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertEqual([], pt['segmentation_labels'])
        self.delete_policy_target(pt['id'], expected_res_status=204)

        labels = []
        pt = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            segmentation_labels=labels)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertItemsEqual([], pt['segmentation_labels'])
        self.delete_policy_target(pt['id'], expected_res_status=204)

        labels = ['red', 'blue']
        pt = self.create_policy_target(
            policy_target_group_id=ptg['id'],
            segmentation_labels=labels)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])

        labels = ['green', 'black', 'red']
        pt = self.update_policy_target(
            pt['id'], segmentation_labels=labels,
            expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])

        labels = []
        pt = self.update_policy_target(
            pt['id'], segmentation_labels=labels,
            expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])

        labels = ['black']
        pt = self.update_policy_target(
            pt['id'], segmentation_labels=labels,
            expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])
        pt = self.show_policy_target(
            pt['id'], expected_res_status=200)['policy_target']
        self.assertItemsEqual(labels, pt['segmentation_labels'])

        self.delete_policy_target(pt['id'], expected_res_status=204)
        session = db_api.get_reader_session()
        rows = (session.query(db.ApicSegmentationLabelDB).filter_by(
                policy_target_id=pt['id']).all())
        self.assertEqual([], rows)


class ExtensionDriverTestCase(test_ext_base.ExtensionDriverTestBase,
                              ExtensionDriverTestCaseMixin):
    _extension_drivers = ['apic_segmentation_label']
    _extension_path = None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys
import time
import traceback

import eventlet
import greenlet
from oslo_service import service as oslo_service

from gbpservice.nfp.core import common as nfp_common
from gbpservice.nfp.core import context
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import watchdog as nfp_watchdog

LOG = nfp_logging.getLogger(__name__)
Service = oslo_service.Service
identify = nfp_common.identify
WATCHDOG = nfp_watchdog.Watchdog

DEFAULT_THREAD_TIMEOUT = (10 * 60)

"""Implements worker process.

    Derives from oslo service.
    Implements the worker functionality.
    Waits for the events from distributor, handles them,
    invokes the registered event handler in a thread.
"""


class NfpWorker(Service):

    def __init__(self, conf, threads=10):
        # REVISIT(mak): Should #threads be a conf ?
        Service.__init__(self, threads=threads)
        # Parent end of duplex pipe
        self.parent_pipe = None
        # Pipe to recv/send messages to distributor
        self.pipe = None
        # Cache of event handlers
        self.controller = None
        self._conf = conf
        self._threads = threads

    def start(self):
        """Service start, runs here till dies.

            When a oslo service is launched, this method
            is invoked.
            Polls for messages from distributor and process
            them.
        """
        # Update the process type in controller.
        self.controller.PROCESS_TYPE = "worker"
        self.controller._pipe = self.pipe
        self.event_handlers = self.controller.get_event_handlers()

        eventlet.spawn_n(self.controller._resending_task)

        while True:
            try:
                event = None
                if self.pipe.poll(0.1):
                    event = self.controller.pipe_recv(self.pipe)
                if event:
                    message = "%s - received event" % (
                        self._log_meta(event))
                    LOG.debug(message)
                    self.controller.decompress(event)
                    self._process_event(event)
            except Exception as e:
                message = "Exception - %s" % (e)
                LOG.error(message)
            # Yeild cpu
            time.sleep(0)

    def _log_meta(self, event=None):
        if event:
            return "(event - %s) - (worker - %d)" % (
                event.identify(), os.getpid())
        else:
            return "(worker - %d)" % (os.getpid())

    def _send_event_ack(self, event):
        # Create new event from existing one
        ack_event = nfp_event.Event(id=event.id)
        ack_event.id = event.id
        desc = nfp_event.EventDesc(**event.desc.__dict__)
        desc.uuid = event.desc.uuid
        desc.flag = nfp_event.EVENT_ACK
        setattr(ack_event, 'desc', desc)
        self.controller.pipe_send(self.pipe, ack_event)

    def _process_event(self, event):
        """Process & dispatch the event.

            Decodes the event type and performs the required
            action.
            Executes the registered event handler in one of the
            thread.
        """
        if event.desc.type == nfp_event.SCHEDULE_EVENT:
            eh, _ = (
                self.event_handlers.get_event_handler(
                    event.id, module=event.desc.target))
            self.dispatch(eh.handle_event, event, eh=eh)
        elif event.desc.type == nfp_event.POLL_EVENT:
            self.dispatch(self._handle_poll_event, event)

    def _repoll(self, ret, event, eh):
        if ret.get('poll', False):
            message = ("(event - %s) - repolling event -"
                       "pending times - %d") % (
                event.identify(), event.desc.poll_desc.max_times)
            LOG.debug(message)
            if event.desc.poll_desc.max_times:
                self.controller.poll_event(
                    event,
                    spacing=event.desc.poll_desc.spacing,
                    max_times=event.desc.poll_desc.max_times)
            else:
                message = ("(event - %s) - max timed out,"
                           "calling event_cancelled") % (event.identify())
                LOG.debug(message)
                eh.event_cancelled(event, 'MAX_TIMED_OUT')

    def _handle_poll_event(self, event):
        ret = {'poll': False}
        event.desc.poll_desc.max_times -= 1
        module = event.desc.target
        poll_handler, _ = (
            self.event_handlers.get_poll_handler(event.id, module=module))
        event_handler, _ = (
            self.event_handlers.get_event_handler(event.id, module=module))
        try:
            try:
                ret = poll_handler(event)
            except TypeError:
                ret = poll_handler(event_handler, event)
            if not ret:
                ret = {'poll': True}
        except greenlet.GreenletExit:
            pass
        except Exception as exc:
            message = "Exception - %r" % (exc)
            LOG.error(message)
            ret = self.dispatch_exception(event_handler, event, exc)
            if not ret:
                ret = {'poll': False}

        self._repoll(ret, event, event_handler)

    def _dispatch(self, handler, event, *args, **kwargs):
        event.context['log_context']['namespace'] = event.desc.target
        context.init(event.context)
        try:
            handler(event, *args)
        except greenlet.GreenletExit:
            self.controller.event_complete(event, result='FAILED')
        except Exception as exc:
            # How to log traceback propery ??
            message = "Exception - %r" % (exc)
            LOG.error(message)
            self.dispatch_exception(kwargs.get('eh'), event, exc)
            self.controller.event_complete(event, result="FAILED")
        finally:
            self._send_event_ack(event)

    def dispatch_exception(self, event_handler, event, exception):
        ret = {}
        try:
            ret = event_handler.handle_exception(event, exception)
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            message = "Traceback: %s" % traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            LOG.error(message)
        finally:
            return ret

    def thread_done(self, th, watchdog=None):
        if watchdog:
            watchdog.cancel()

    def thread_timedout(self, thread=None):
        if thread:
            eventlet.greenthread.kill(thread.thread)

    def dispatch(self, handler, event, *args, **kwargs):
        if self._threads:
            th = self.tg.add_thread(
                self._dispatch, handler, event, *args, **kwargs)
            message = "%s - (handler - %s) - dispatched to thread " % (
                self._log_meta(), identify(handler))
            LOG.debug(message)
            wd = WATCHDOG(self.thread_timedout,
                          seconds=DEFAULT_THREAD_TIMEOUT, thread=th)
            th.link(self.thread_done, watchdog=wd)
        else:
            try:
                handler(event, *args)
                message = "%s - (handler - %s) - invoked" % (
                    self._log_meta(), identify(handler))
                LOG.debug(message)
                self._send_event_ack(event)
            except Exception as exc:
                message = "Exception from module's event handler - %s" % (exc)
                LOG.error(message)
                self.dispatch_exception(kwargs.get('eh'), event, exc)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exceptions used by NodeCompositionPlugin and drivers."""

from neutron_lib import exceptions

from gbpservice._i18n import _


class NodeDriverError(exceptions.NeutronException):
    """Node driver call failed."""
    message = _("%(method)s failed.")


class NodeCompositionPluginException(exceptions.NeutronException):
    """Base for node driver exceptions returned to user."""
    pass


class PlumbingException(exceptions.NeutronException):
    """Base for node driver exceptions returned to user."""
    pass


class NodeCompositionPluginBadRequest(exceptions.BadRequest,
                                      NodeCompositionPluginException):
    """Base for node driver bad request exceptions returned to user."""
    pass


class OneSpecPerInstanceAllowed(NodeCompositionPluginBadRequest):
    message = _("The Node Composition Plugin only supports one Servicechain"
                "Spec per Servicechain Instance.")


class NoDriverAvailableForAction(NodeCompositionPluginBadRequest):
    message = _("The Node Composition Plugin can't find any Node Driver "
                "available for executing %(action)s on node %(node_id)s. "
                "This may be caused by a Servicechain Node misconfiguration "
                "or an unsupported Service Profile.")


class ServiceProfileInUseByAnInstance(NodeCompositionPluginBadRequest):
    message = _("Cannot update Service Profile %(profile_id)s because it's "
                "used by servicechain instance %(instance_id)s.")


class NotAvailablePTGForTargetRequest(PlumbingException):
    message = _("PTG of type %(ptg_type)s doesn't exist for service chain "
                "instance %(instance)s. However, it is required by the "
                "scheduled Node Driver in order to deploy Node %(node)s")


class InuseSpecNodeUpdateNotAllowed(NodeCompositionPluginBadRequest):
    message = _("The Node Composition Plugin does not support updating the "
                "nodes in an instantiated servicechain spec.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nfd_cluster_mapping_info

Revision ID: ef5a69e5bcc5
Revises: fce38a8588a2
Create Date: 2016-11-23 21:17:53.858242

"""


# revision identifiers, used by Alembic.
revision = 'ef5a69e5bcc5'
down_revision = 'fce38a8588a2'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'nfd_cluster_mapping_info',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('network_function_device_id',
                  sa.String(length=36), nullable=True),
        sa.Column('cluster_group', sa.Integer(), nullable=True),
        sa.Column('virtual_ip', sa.String(length=36), nullable=True),
        sa.Column('multicast_ip', sa.String(length=36), nullable=True),
        sa.Column('cluster_name', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['network_function_device_id'],
                                ['nfp_network_function_devices.id'],
                                ondelete='SET NULL'),
        sa.PrimaryKeyConstraint('id')
    )

    op.add_column(
        'nfp_network_function_devices',
        sa.Column('provider_metadata', sa.String(length=1024), nullable=True)
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE: The purpose of this module is to provide a nop test to verify that
# the functional gate is working.

from neutron.tests.functional import base as functional_base


class TestPolicyMappingDriver(functional_base.BaseSudoTestCase):

    """Test policy mapping driver."""

    # NOTE: Tests may be added/removed/changed, when this is fleshed out
    # in future commits.

    def test_policy_target_group_create(self):
        """Test PTG creation."""
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import rpc as n_rpc
from oslo_config import cfg
import oslo_messaging as messaging

from gbpservice.contrib.nfp.configurator.lib import constants as const
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

n_rpc.init(cfg.CONF)

LOG = nfp_logging.getLogger(__name__)


class AgentBaseRPCManager(object):
    """Implements base class for all service agents.

    Common methods for service agents are implemented in this class.
    Configurator module invokes these methods through the service
    agent's child class instance.

    """

    def __init__(self, sc, conf):
        self.sc = sc
        self.conf = conf

    def validate_request(self, sa_req_list, notification_data):
        """Preliminary validation of function input.

        :param sa_req_list: List of data blobs prepared by de-multiplexer
        for service agents processing.
        :param notification_data: Notification blobs prepared by the service
        agents after processing requests blobs. Each request blob will have
        a corresponding notification blob.

        Returns: True if validation passes. False if validation fails.

        """

        if (isinstance(sa_req_list, list) and
                isinstance(notification_data, dict)):
            return True
        else:
            return False

    def get_diff_of_dict(self, old_dict, new_dict):
        """Getting difference between two dict.

        :param Two dictionary

        Returns: Two dictionary which has different values for same keys.

        """
        diff_values = []
        new_val = {}
        old_val = {}
        for key in new_dict:
            if old_dict.get(key) != new_dict.get(key):
                diff_values.append(key)

        for value in diff_values:
            if value == 'description':
                pass
            else:
                new_val[value] = new_dict.get(value)
                old_val[value] = old_dict.get(value)
        return old_val, new_val

    def process_request(self, sa_req_list, notification_data):
        """Forwards the RPC message from configurator to service agents.

        Checks if the request message contains multiple data blobs. If multiple
        data blobs are found, a batch event is generated otherwise a single
        event.

        :param sa_req_list: List of data blobs prepared by de-multiplexer
        for service agents processing.
        :param notification_data: Notification blobs prepared by the service
        agents after processing requests blobs. Each request blob will have
        a corresponding notification blob.

        Returns: None

        """

        # In case of malformed input, send failure notification
        if not self.validate_request(sa_req_list, notification_data):
            # REVISIT(JAGADISH): Need to send failure notification
            return

        # Multiple request data blobs needs batch processing. Send batch
        # processing event or do direct processing of single request data blob
        if (len(sa_req_list) > 1):
            LOG.info("Creating event PROCESS BATCH")
            args_dict = {
                'sa_req_list': sa_req_list,
                'notification_data': notification_data
            }
            ev = self.sc.new_event(id=const.EVENT_PROCESS_BATCH,
                                   data=args_dict, key=None)
            self.sc.post_event(ev)
        else:
            agent_info = sa_req_list[0]['agent_info']
            # Renaming the neutron context in resource data of *aaS to context.
            # Adding agent_info which contains information required for
            # demux and response data in agent to neutron_context in *aaS
            if not sa_req_list[0]['is_generic_config'] and not (
                    agent_info['resource'] in const.NFP_SERVICE_LIST):
                # Here, the neutron context is overloaded with agent_info
                # dict which contains the API context in addition to other
                # fields like service type, service vendor, resource etc.
                # The agent_info dict is constructed inside the demuxer library
                sa_req_list[0]['resource_data']['neutron_context'].update(
                    {'agent_info': agent_info})
                # When calling the *aaS or NFPService agents, the
                # "neutron context" passed inside the resource data is
                # renamed to "context"
                sa_req_list[0]['resource_data']['context'] = sa_req_list[0][
                    'resource_data'].pop('neutron_context')
                getattr(self, sa_req_list[0]['method'])(
                    **sa_req_list[0]['resource_data'])
            else:
                sa_req_list[0]['agent_info'].update(
                    {'notification_data': notification_data})
                getattr(self, sa_req_list[0]['method'])(
                    agent_info, sa_req_list[0]['resource_data'])


class AgentBaseNotification(object):
    """Enqueues notification event into notification queue

    Responses from the REST calls made to the VM are fed to under the
    cloud components using this notification handle.
    """

    API_VERSION = '1.0'

    def __init__(self, sc):
        self.sc = sc
        self.topic = const.NOTIFICATION_QUEUE
        target = messaging.Target(topic=self.topic,
                                  version=self.API_VERSION)
        self.client = n_rpc.get_client(target)
        self.cctxt = self.client.prepare(version=self.API_VERSION,
                                         topic=self.topic)

    def _notification(self, data):
        """Enqueues notification event into const.NOTIFICATION_QUEUE

        These events are enqueued into notification queue and are retrieved
        when get_notifications() API lands on configurator.

        :param data: Event data blob

        Returns: None

        """
        self.cctxt.cast(self, 'send_notification', notification_data=[data])

    def to_dict(self):
        return {}


class AgentBaseEventHandler(nfp_api.NfpEventHandler):
    """ Super class for all agents to handle batch events.

    """

    def __init__(self, sc, drivers, rpcmgr):
        self.sc = sc
        self.drivers = drivers
        self.rpcmgr = rpcmgr
        self.notify = AgentBaseNotification(self.sc)

    def process_batch(self, ev):
        """Processes a request with multiple data blobs.

        Configurator processes the request with multiple data blobs and sends
        a list of service information to be processed. This function goes
        through the list of service information and invokes specific service
        driver methods. After processing each request data blob, notification
        data blob is prepared.

        :param ev: Event instance that contains information of event type and
        corresponding event data to be processed.

        """

        # Get service agent information list and notification data list
        # from the event data
        sa_req_list = ev.data.get('sa_req_list')
        notification_data = ev.data.get('notification_data')

        for request in sa_req_list:
            try:
                # Process the first data blob from the request list.
                # Get necessary parameters needed for driver method invocation.
                method = request['method']
                is_generic_config = request['is_generic_config']
                resource_data = request['resource_data']
                agent_info = request['agent_info']
                resource = agent_info['resource']
                # agent_info contains the API context.
                context = agent_info['context']
                service_vendor = agent_info['service_vendor']
                service_type = agent_info['resource_type']
                service_feature = agent_info['service_feature']
                if not is_generic_config:
                    sa_req_list[0]['resource_data']['context'] = sa_req_list[
                        0]['resource_data'].pop('neutron_context')

                # Get the service driver and invoke its method
                driver = self._get_driver(service_type, service_vendor,
                                          service_feature)

                # Service driver should return "success" on successful API
                # processing. All other return values and exceptions are
                # treated as failures.
                if is_generic_config:
                    result = getattr(driver, method)(context, resource_data)
                else:
                    result = getattr(driver, method)(**resource_data)
                success = True if result == 'SUCCESS' else False
            except Exception as err:
                result = ("Failed to process %s request. %s" %
                          (method, str(err).capitalize()))
                success = False
            finally:
                # Prepare success notification and populate notification
                # data list
                if result in const.SUCCESS:
                    data = {'status_code': const.SUCCESS}
                else:
                    data = {'status_code': const.FAILURE,
                            'error_msg': result}

                msg = {'info': {'service_type': service_type,
                                'context': context},
                       'notification': [{'resource': resource,
                                         'data': data}]
                       }
                # If the data processed is first one, then prepare notification
                # dict. Otherwise, append the notification to the kwargs list.
                # Whether it is a data batch or single data blob request,
                # notification generated will be single dictionary. In case of
                # batch, multiple notifications are sent in the kwargs list.
                if not notification_data:
                    notification_data.update(msg)
                else:
                    data = {'resource': resource,
                            'data': data}
                    notification_data['notification'].append(data)

            if not success:
                self.notify._notification(notification_data)
                raise Exception(msg)

        self.notify._notification(notification_data)


def init_agent_complete(cm, sc, conf):
    """Placeholder method to satisfy configurator module agent loading."""
    pass


def init_agent(cm, sc, conf):
    """Placeholder method to satisfy configurator module agent loading."""
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import zlib

from oslo_log import log as logging
import oslo_serialization.jsonutils as jsonutils

from pecan.hooks import HookController
from pecan.hooks import PecanHook
from pecan import rest

LOG = logging.getLogger(__name__)


class ZipperHook(PecanHook):

    def before(self, state):
        if state.request.method.upper() != 'GET':
            try:
                zippedBody = state.request.body
                body = zlib.decompress(zippedBody)
                body = jsonutils.loads(body)
                state.request.json_body = body
                state.request.content_type = "application/json"
            except Exception as e:
                msg = ("Failed to process data ,Reason: %s" % (e))
                LOG.error(msg)

    def after(self, state):
        data = state.response.body
        state.response.body = zlib.compress(data)
        state.response.content_type = "application/octet-stream"


class BaseController(rest.RestController, HookController):
    """This is root controller that forward the request to __init__.py
    file inside controller folder inside v1

    """
    __hooks__ = [ZipperHook()]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

API_NAME = 'sfc_patch'
API_VERSION_OPTION = ''


def build_option_parser(parser):
    """Hook to add global options"""
    return parser


try:
    from networking_sfc.cli import port_pair_group
    from networking_sfc.osc.sfc import port_pair_group as osc_ppg

    # Removing field limitation
    def add_known_arguments(self, parser):
        parser.add_argument(
            'name',
            metavar='NAME',
            help=port_pair_group._('Name of the Port Pair Group.'))
        port_pair_group.add_common_arguments(parser)
        parser.add_argument(
            '--port-pair-group-parameters',
            type=port_pair_group.utils.str2dict,
            help=port_pair_group._(
                'Dictionary of Port pair group parameters. '))

    def get_parser(self, prog_name):
        parser = super(osc_ppg.CreatePortPairGroup, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='NAME',
            help=osc_ppg._('Name of the Port Pair Group.'))
        parser.add_argument(
            '--description',
            help=osc_ppg._('Description for the Port Pair Group.'))
        parser.add_argument(
            '--port-pair',
            metavar='PORT-PAIR',
            dest='port_pairs',
            default=[],
            action='append',
            help=osc_ppg._('ID or name of the Port Pair.'
                           'This option can be repeated.'))
        parser.add_argument(
            '--port-pair-group-parameters',
            type=osc_ppg.nc_utils.str2dict,
            help=osc_ppg._('Dictionary of Port pair group parameters. '))
        return parser

    port_pair_group.PortPairGroupCreate.add_known_arguments = (
        add_known_arguments)
    osc_ppg.CreatePortPairGroup.get_parser = get_parser
except ImportError:
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import httplib
import socket
import zlib

import exceptions
import httplib2
from oslo_serialization import jsonutils
import six.moves.urllib.parse as urlparse

from gbpservice._i18n import _
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


class RestClientException(exceptions.Exception):

    """ RestClient Exception """


class UnixHTTPConnection(httplib.HTTPConnection):

    """Connection class for HTTP over UNIX domain socket."""

    def __init__(self, host, port=None, strict=None, timeout=None,
                 proxy_info=None):
        httplib.HTTPConnection.__init__(self, host, port, strict)
        self.timeout = timeout
        self.socket_path = '/var/run/uds_socket'

    def connect(self):
        """Method used to connect socket server."""
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        if self.timeout:
            self.sock.settimeout(self.timeout)
        try:
            self.sock.connect(self.socket_path)
        except socket.error as exc:
            raise RestClientException(
                _("Caught exception socket.error : %s") % exc)


class UnixRestClient(object):

    def _http_request(self, url, method_type, headers=None, body=None):
        try:
            h = httplib2.Http()
            resp, content = h.request(
                url,
                method=method_type,
                headers=headers,
                body=body,
                connection_type=UnixHTTPConnection)
            return resp, content

        except httplib2.ServerNotFoundError:
            raise RestClientException(_("Server Not Found"))

        except exceptions.Exception as e:
            raise RestClientException(_("httplib response error %s") % (e))

    def send_request(self, path, method_type, request_method='http',
                     server_addr='127.0.0.1',
                     headers=None, body=None):
        """Implementation for common interface for all unix crud requests.
        Return:Http Response
        """
        # prepares path, body, url for sending unix request.
        if method_type.upper() != 'GET':
            body = jsonutils.dumps(body)
            body = zlib.compress(body)

        path = '/v1/nfp/' + path
        url = urlparse.urlunsplit((
            request_method,
            server_addr,
            path,
            None,
            ''))

        try:
            resp, content = self._http_request(url, method_type,
                                               headers=headers, body=body)
            if content != '':
                content = zlib.decompress(content)
            message = "%s:%s" % (resp, content)
            LOG.info(message)
        except RestClientException as rce:
            message = "ERROR : %s" % (rce)
            LOG.error(message)
            raise rce

        success_code = [200, 201, 202, 204]
        # Evaluate responses into success and failures.
        # Raise exception for failure cases which needs
        # to be handled by caller.
        if success_code.__contains__(resp.status):
            return resp, content
        elif resp.status == 400:
            raise RestClientException(_("HTTPBadRequest: %s") % resp.reason)
        elif resp.status == 401:
            raise RestClientException(_("HTTPUnauthorized: %s") % resp.reason)
        elif resp.status == 403:
            raise RestClientException(_("HTTPForbidden: %s") % resp.reason)
        elif resp.status == 404:
            raise RestClientException(_("HttpNotFound: %s") % resp.reason)
        elif resp.status == 405:
            raise RestClientException(
                _("HTTPMethodNotAllowed: %s") % resp.reason)
        elif resp.status == 406:
            raise RestClientException(_("HTTPNotAcceptable: %s") % resp.reason)
        elif resp.status == 408:
            raise RestClientException(
                _("HTTPRequestTimeout: %s") % resp.reason)
        elif resp.status == 409:
            raise RestClientException(_("HTTPConflict: %s") % resp.reason)
        elif resp.status == 415:
            raise RestClientException(
                _("HTTPUnsupportedMediaType: %s") % resp.reason)
        elif resp.status == 417:
            raise RestClientException(
                _("HTTPExpectationFailed: %s") % resp.reason)
        elif resp.status == 500:
            raise RestClientException(_("HTTPServerError: %s") % resp.reason)
        else:
            raise Exception(_('Unhandled Exception code: %(st)s %(reason)s') %
                            {'st': resp.status, 'reason': resp.reason})


def get(path):
    """Implements get method for unix restclient
    Return:Http Response
    """
    return UnixRestClient().send_request(path, 'GET')


def put(path, body):
    """Implements put method for unix restclient
    Return:Http Response
    """
    headers = {'content-type': 'application/octet-stream'}
    return UnixRestClient().send_request(
        path, 'PUT', headers=headers, body=body)


def post(path, body, delete=False):
    """Implements post method for unix restclient
    Return:Http Response
    """
    # Method-Type added here,as DELETE/CREATE
    # both case are handled by post as delete also needs
    # to send data to the rest-unix-server.
    headers = {'content-type': 'application/octet-stream'}
    if delete:
        headers.update({'method-type': 'DELETE'})
    else:
        headers.update({'method-type': 'CREATE'})
    return UnixRestClient().send_request(
        path, 'POST', headers=headers, body=body)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FakeObjects(object):
    """ Implements fake objects for assertion.

    """

    sc = 'sc'
    conf = 'conf'
    context = 'APIcontext'
    kwargs = {'vmid': 'vmid'}
    rpcmgr = 'rpcmgr'
    drivers = 'drivers'


class FakeEventNfpService(object):
    """ Implements a fake event class for generic config for
        process framework to use

    """

    def __init__(self):
        self.data = {
                    'context': {
                            'resource': 'heat',
                            'notification_data': {},
                            'resource_type': 'firewall',
                            'service_vendor': 'vyos',
                            'context': 'APIcontext'},
                    'resource_data': 'some data'}
        self.id = 'dummy'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg as oslo_cfg
from oslo_log import log as oslo_logging

oslo_logging.register_options(oslo_cfg.CONF)


class Object(object):
    pass


def init():
    """Initialize logging. """
    product_name = "nfp"
    oslo_logging.setup(oslo_cfg.CONF, product_name)


def _is_class(obj):
    return 'class' in str(type(obj))


def _name(obj):
    """Helper method to construct name of an object.

    'module.class' if object is of type 'class'
    'module.class.method' if object is of type 'method'
    """
    # If it is callable, then it is a method
    if callable(obj):
        return "{0}.{1}.{2}".format(
            type(obj.im_self).__module__,
            type(obj.im_self).__name__,
            obj.__name__)
    # If obj is of type class
    elif _is_class(obj):
        return "{0}.{1}".format(
            type(obj).__module__,
            type(obj).__name__)
    else:
        return obj.__name__


def identify(obj):
    """Helper method to display identity an object.

    Useful for logging. Decodes based on the type of obj.
    Supports 'class' & 'method' types for now.

    :param obj: Object (Class/Method supported.)
    Returns: String. Identification of the object.
    """
    prefix = obj._NAME_ if hasattr(obj, '_NAME_') else ''
    try:
        return "([%s] %s)" % (prefix, _name(obj))
    except Exception:
        # Some unknown type, returning empty
        return ""
